package com.massa.util.extensions.nativeproperty;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.concurrent.atomic.AtomicInteger;

import com.massa.log.Log;
import com.massa.log.LogFactory;
import com.massa.util.MessageCode;
import com.massa.util.MessageInfo;
import com.massa.util.UtilsException;
import com.massa.util.extensions.nativeproperty.compiler.InMemoryJavaCompiler;
import com.massa.util.property.IArrayProvider;
import com.massa.util.property.IFieldAccessor;
import com.massa.util.property.IMethodInvoker;
import com.massa.util.property.ReflectionFieldAccessor;
import com.massa.util.property.ReflectionMethodInvoker;
import com.massa.util.property.ReflectionPropertyAccessCache;

/**
 * Overrides reflection access cache in order to provide dynamically generated accessors and invokers.
 */
public class NativePropertyAccessCache extends ReflectionPropertyAccessCache {
	private static final Log LOG = LogFactory.getLog(NativePropertyAccessCache.class);

	private static final NativePropertyAccessCache INSTANCE = new NativePropertyAccessCache();

	public static final AtomicInteger IDX = new AtomicInteger();

	/**
	 * Protected as this class should be used as Singleton but might be extended.
	 */
	protected NativePropertyAccessCache() {
	}

	/**
	 * @return Default global instance.
	 */
	public static NativePropertyAccessCache getInstance() {
		return NativePropertyAccessCache.INSTANCE;
	}

	/**
	 * Builds specific {@link IMethodInvoker} implementation instead of {@link ReflectionMethodInvoker}.
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected IMethodInvoker buildMethodInvoker(final Method m) {
		if (Modifier.isPublic(m.getModifiers()) && Modifier.isPublic(m.getDeclaringClass().getModifiers())) {
			final int idx = NativePropertyAccessCache.IDX.getAndAdd(1);
			final String clazzName = "$MRulesNativeAccess$"+idx;
			final String clazzFullName = "com.massa.util.extensions.nativeproperty." + clazzName;
			Class<IMethodInvoker> clazz = null;
			try {
				final StringBuilder sourceCode = NativePropertyAccessCache.buildMethodSourceCode(m, clazzName);
				clazz = (Class<IMethodInvoker>) InMemoryJavaCompiler.compile(clazzFullName, sourceCode.toString(), NativePropertyAccessCache.getOptions());
			} catch (final UtilsException e) {
				if (NativePropertyAccessCache.LOG.isWarnEnabled()) {
					NativePropertyAccessCache.LOG.warn("Generated class " + clazzFullName + " could not be created. Defaulting to reflection.", e);
				}
			}
			if (clazz != null) {
				try {
					final Constructor<IMethodInvoker> constr = clazz.getConstructor(Method.class);
					final IMethodInvoker inst = constr.newInstance(m);
					if (NativePropertyAccessCache.LOG.isDebugEnabled()) {
						NativePropertyAccessCache.LOG.debug("Generated class " + clazzFullName + " was created and instantiated for Method " + m + '.');
					}
					return inst;
				} catch (final Exception e) {
					if (NativePropertyAccessCache.LOG.isWarnEnabled()) {
						NativePropertyAccessCache.LOG.warn("Generated class " + clazzFullName + " could not be instantiated. Defaulting to reflection.", e);
					}
				}
			}
		} else {
			if (NativePropertyAccessCache.LOG.isDebugEnabled()) {
				NativePropertyAccessCache.LOG.debug("Method " + m + " is not accessible. Defaulting to reflection.");
			}
		}
		return new ReflectionMethodInvoker(m);
	}

	/**
	 * Builds specific {@link IFieldAccessor} implementation instead of {@link ReflectionFieldAccessor}.
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected IFieldAccessor buildFieldAccessor(final Field f) {
		if (Modifier.isPublic(f.getModifiers())) {
			final int idx = NativePropertyAccessCache.IDX.getAndAdd(1);
			final String clazzName = "$MRulesNativeAccess$"+idx;
			final String clazzFullName = "com.massa.util.extensions.nativeproperty." + clazzName;
			Class<IFieldAccessor> clazz = null;
			try {
				final StringBuilder sourceCode = NativePropertyAccessCache.buildFieldSourceCode(f, clazzName);

				clazz = (Class<IFieldAccessor>) InMemoryJavaCompiler.compile(clazzFullName, sourceCode.toString(), NativePropertyAccessCache.getOptions());
			} catch (final UtilsException e) {
				if (NativePropertyAccessCache.LOG.isWarnEnabled()) {
					NativePropertyAccessCache.LOG.warn("Generated class " + clazzFullName + " could not be created. Defaulting to reflection.", e);
				}
			}
			if (clazz != null) {
				try {
					final Constructor<IFieldAccessor> constr = clazz.getConstructor(Field.class);
					final IFieldAccessor inst = constr.newInstance(f);
					if (NativePropertyAccessCache.LOG.isDebugEnabled()) {
						NativePropertyAccessCache.LOG.debug("Generated class " + clazzFullName + " was created and instantiated for Field " + f + '.');
					}
					return inst;
				} catch (final Exception e) {
					if (NativePropertyAccessCache.LOG.isWarnEnabled()) {
						NativePropertyAccessCache.LOG.warn("Generated class " + clazzFullName + " could not be instantiated. Defaulting to reflection.", e);
					}
				}
			}
		} else {
			if (NativePropertyAccessCache.LOG.isDebugEnabled()) {
				NativePropertyAccessCache.LOG.debug("Field " + f + " is not accessible. Defaulting to reflection.");
			}
		}
		return new ReflectionFieldAccessor(f);
	}

	private static StringBuilder buildFieldSourceCode(final Field f, final String clazzName) {
		final StringBuilder sourceCode = new StringBuilder(1000);
		sourceCode.append("package com.massa.util.extensions.nativeproperty;\n");
		sourceCode.append("public class ").append(clazzName).append(" extends ").append(NativePropertyAccessCache.getInvokeName(ReflectionFieldAccessor.class)).append(" {\n");
		sourceCode.append("	public ").append(clazzName).append("(final java.lang.reflect.Field f) {\n");
		sourceCode.append("		super(f);\n");
		sourceCode.append("	}\n");
		sourceCode.append("	@SuppressWarnings(\"unchecked\")");
		sourceCode.append("	@Override");
		sourceCode.append("	public Object get(Object on) throws ").append(NativePropertyAccessCache.getInvokeName(UtilsException.class)).append(" {\n");
		sourceCode.append("		return ((").append(NativePropertyAccessCache.getInvokeName(f.getDeclaringClass())).append(") on).").append(f.getName()).append(";\n");
		sourceCode.append("	}\n");
		sourceCode.append("	@SuppressWarnings(\"unchecked\")");
		sourceCode.append("	@Override");
		sourceCode.append("	public void set(Object on, Object value) throws ").append(NativePropertyAccessCache.getInvokeName(UtilsException.class)).append(" {\n");
		sourceCode.append("		((").append(NativePropertyAccessCache.getInvokeName(f.getDeclaringClass())).append(") on).").append(f.getName()).append(" = (").append(NativePropertyAccessCache.getInvokeName(f.getType())).append(") value;\n");
		sourceCode.append("	}\n");
		sourceCode.append("}\n");
		return sourceCode;
	}

	private static StringBuilder buildMethodSourceCode(final Method m, final String clazzName) {
		final Class<?>[] params = m.getParameterTypes();
		final StringBuilder sourceCode = new StringBuilder(1000);
		sourceCode.append("package com.massa.util.extensions.nativeproperty;\n");
		sourceCode.append("public class ").append(clazzName).append(" extends ").append(NativePropertyAccessCache.getInvokeName(ReflectionMethodInvoker.class)).append(" {\n");
		sourceCode.append("	public ").append(clazzName).append("(final java.lang.reflect.Method m) {\n");
		sourceCode.append("		super(m);\n");
		sourceCode.append("	}\n");

		sourceCode.append("	@SuppressWarnings(\"unchecked\")");
		sourceCode.append("	@Override");
		sourceCode.append("	public Object invoke(final ").append(NativePropertyAccessCache.getInvokeName(IArrayProvider.class)).append(" ap, Object on) throws ").append(NativePropertyAccessCache.getInvokeName(UtilsException.class)).append(" {\n");
		if (params.length == 0) {
			NativePropertyAccessCache.appendCallReturn(sourceCode, m, params, false);
		} else {
			NativePropertyAccessCache.appendThrow(sourceCode);
		}
		sourceCode.append("	}\n");

		sourceCode.append("	@SuppressWarnings(\"unchecked\")");
		sourceCode.append("	@Override");
		sourceCode.append("	public Object invoke(final ").append(NativePropertyAccessCache.getInvokeName(IArrayProvider.class)).append(" ap, Object on, Object arg0) throws ").append(NativePropertyAccessCache.getInvokeName(UtilsException.class)).append(" {\n");
		if (params.length == 1) {
			NativePropertyAccessCache.appendCallReturn(sourceCode, m, params, false);
		} else {
			NativePropertyAccessCache.appendThrow(sourceCode);
		}
		sourceCode.append("	}\n");

		sourceCode.append("	@SuppressWarnings(\"unchecked\")");
		sourceCode.append("	@Override");
		sourceCode.append("	public Object invoke(final ").append(NativePropertyAccessCache.getInvokeName(IArrayProvider.class)).append(" ap, Object on, Object arg0, Object arg1) throws ").append(NativePropertyAccessCache.getInvokeName(UtilsException.class)).append(" {\n");
		if (params.length == 2) {
			NativePropertyAccessCache.appendCallReturn(sourceCode, m, params, false);
		} else {
			NativePropertyAccessCache.appendThrow(sourceCode);
		}
		sourceCode.append("	}\n");

		sourceCode.append("	@SuppressWarnings(\"unchecked\")");
		sourceCode.append("	@Override");
		sourceCode.append("	public Object invoke(final ").append(NativePropertyAccessCache.getInvokeName(IArrayProvider.class)).append(" ap, final Object on, final Object ... arguments) throws ").append(NativePropertyAccessCache.getInvokeName(UtilsException.class)).append(" {\n");

		if (params.length == 0) {
			sourceCode.append("		if (arguments != null && arguments.length > 0) {\n");
			NativePropertyAccessCache.appendThrow(sourceCode);
		} else {
			sourceCode.append("		if (arguments == null || arguments.length != ").append(params.length).append(") {\n");
			NativePropertyAccessCache.appendThrow(sourceCode);
		}
		sourceCode.append("		}\n");

		NativePropertyAccessCache.appendCallReturn(sourceCode, m, params, true);
		sourceCode.append("	}\n");

		sourceCode.append("}\n");
		return sourceCode;
	}

	private static void appendCallReturn(final StringBuilder sourceCode, final Method m, final Class<?>[] params, final boolean array) {
		if (void.class.equals(m.getReturnType())) {
			sourceCode.append("		");
			NativePropertyAccessCache.appendCall(sourceCode, m, params, array);
			sourceCode.append("		return null;\n");
		} else {
			sourceCode.append("		return ");
			NativePropertyAccessCache.appendCall(sourceCode, m, params, array);
		}
	}

	private static void appendThrow(final StringBuilder sourceCode) {
		sourceCode.append("			throw new ").append(NativePropertyAccessCache.getInvokeName(UtilsException.class)).append("(new ")
		.append(NativePropertyAccessCache.getInvokeName(MessageInfo.class)).append("(new ")
		.append(NativePropertyAccessCache.getInvokeName(MessageCode.DynamicMessageCode.class))
		.append("(\"ADD_NATIVE_INVOCATION\", \"Incorrect number of arguments for invocation.\")));\n");
	}

	private static void appendCall(final StringBuilder sourceCode, final Method m, final Class<?>[] params, final boolean array) {
		sourceCode.append("((").append(NativePropertyAccessCache.getInvokeName(m.getDeclaringClass())).append(") on).").append(m.getName()).append("(\n");
		for (int i = 0 ; i < params.length ; i++) {
			sourceCode.append("			(").append(NativePropertyAccessCache.getInvokeName(params[i])).append(')');
			if (array) {
				sourceCode.append("arguments[").append(i).append(']');
			} else {
				sourceCode.append("arg").append(i);
			}
			if (i > 0) {
				sourceCode.append(',');
			}
			sourceCode.append('\n');
		}
		sourceCode.append("		);\n");
	}

	private static String getInvokeName(final Class<?> c) {
		return c.getCanonicalName();
	}

	private static  ArrayList<String> getOptions() {
		final ArrayList<String> options = new ArrayList<String>();
		options.add("-g:none");
		options.add("-nowarn");
		return options;
	}

}
