package com.massa.util.extensions.nativeproperty.compiler;

import java.net.URISyntaxException;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Arrays;

import javax.tools.JavaCompiler;
import javax.tools.JavaFileObject;
import javax.tools.StandardJavaFileManager;
import javax.tools.ToolProvider;

import com.massa.log.Log;
import com.massa.log.LogFactory;
import com.massa.util.MessageInfo;
import com.massa.util.UtilsException;
import com.massa.util.extensions.nativeproperty.Messages;

/**
 * Allows to dynamically compile Java code.
 */
public final class InMemoryJavaCompiler {
	private static final Log LOG = LogFactory.getLog(InMemoryJavaCompiler.class);

	private static final JavaCompiler JAVAC = ToolProvider.getSystemJavaCompiler();

	private InMemoryJavaCompiler() {
	}

	/**
	 * Launches Java Code compilation.
	 * @return newly created class Object.
	 */
	public static Class<?> compile(final String className, final String sourceCodeInText, final Iterable<String> options) throws UtilsException {
		try {
			if (InMemoryJavaCompiler.LOG.isTraceEnabled()) {
				InMemoryJavaCompiler.LOG.trace("Compiling new class:" + className + ". Source code: \n" + sourceCodeInText);
			} else if (InMemoryJavaCompiler.LOG.isDebugEnabled()) {
				InMemoryJavaCompiler.LOG.debug("Compiling new class:" + className + '.');
			}
			final SourceCode sourceCode = new SourceCode(className, sourceCodeInText);
			final CompiledCode compiledCode = new CompiledCode(className);
			final Iterable<? extends JavaFileObject> compilationUnits = Arrays.asList(sourceCode);
			final DynamicClassLoader cl = AccessController.doPrivileged(new PrivilegedAction<DynamicClassLoader>() {
				@Override
				public DynamicClassLoader run() {
					return new DynamicClassLoader(ClassLoader.getSystemClassLoader());
				}
			});
			final StandardJavaFileManager standardFileManager = InMemoryJavaCompiler.JAVAC.getStandardFileManager(null, null, null);
			final ExtendedStandardJavaFileManager fileManager = new ExtendedStandardJavaFileManager(standardFileManager, compiledCode, cl);
			final JavaCompiler.CompilationTask task = InMemoryJavaCompiler.JAVAC.getTask(null, fileManager, null, options, null, compilationUnits);
			if (task.call()) {
				return cl.loadClass(className);
			}
			throw new UtilsException(new MessageInfo(Messages.ADD_NATIVE_COMPILATION_FAILED));
		} catch (final URISyntaxException e) {
			throw new UtilsException(new MessageInfo(Messages.ADD_NATIVE_COMPILATION_FAILED), e);
		} catch (final ClassNotFoundException e) {
			throw new UtilsException(new MessageInfo(Messages.ADD_NATIVE_COMPILATION_FAILED), e);
		}
	}
}
